/** Utilities for GPilotS
  @file
  @author C. Claveleira
  @date 1999-2004
  
*/

/* 
   Copyright (C) 1999-2004 Christian Claveleira (Christian.Claveleira@cru.fr)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   
 */

static char rcsid[] = "$Id: gpsutil.c,v 1.11 2004/11/17 21:15:01 clavelei Exp $";

#include <PalmOS.h>
#include <CharAttr.h>
#include <stdio.h>
#include "MathLib.h"
//#include "gpslib.h"
#include "Garmin.h"
#include "types.h"
#include "log.h"
#include "gpsutil.h"
#include <TxtGlue.h>
//#include "MathLib.c"

#define min(a, b) (((a) < (b)) ? (a) : (b))
#undef TxtGlueCharIsDigit  // TxtGlueCharIsDigit is buggy on some PalmOS
#define TxtGlueCharIsDigit(c) (((c)<='9') && ((c)>='0'))

int DebugLevel;
int DebugIndentLevel;
static LogProcPtr LogOutput;
void DebugPrintf (char * formatStr, ...);

void
SetMathLibRef (UInt16 ref)
{
    MathLibRef = ref;
}



			  // *********   String related functions :


/**
   Concatenation of a string and a char array.
 The array don't need to be null terminated and, in all cases, the resulting
 string is null terminated (contrary to StrNCat)

@param dstp null terminated destination string
@param srcp source string/array to catenate to dstp
@param  dstsize allocated size of the destination string
@param  srcsize number of chars to copy
@return nothing
*/

void
StrNCatSafe (char *dstp, char *srcp, UInt16 dstsize, UInt16 srcsize)
{
    debug (GPSUDBG_CALLS, ("<->StrNCatSafe(dstsize=%d, srcsize=%d", dstsize, srcsize));
    if (dstsize > (StrLen (dstp) + srcsize))
	StrNCat (dstp, srcp, StrLen (dstp) + srcsize + 1);
    else
	StrNCat (dstp, srcp, dstsize - 1);
    dstp[dstsize - 1] = 0;
}							   // StrNCatSafe


/**
   Copy at most dstsize-1 bytes from source to destination
string. Ensures that the destination string is null terminated
(dstsize'th byte is 0) except if dstsize=0

   @param dstp pointer on destination string
   @param srcp pointer on source string
   @param  dstsize size to fill (including null)
   @return addr of next char after last char copied
*/

char *
StrNCopySafe (char *dstp, char *srcp, UInt16 dstsize)
{
    debug (GPSUDBG_CALLS, ("StrNCopySafe(dstsize=%d)", dstsize));
    if (dstsize)
      {
	  for (; dstsize; dstsize--)
	      if (!(*dstp++ = *srcp++))
		  break;
	  *(dstp - 1) = 0;
      }
    return (dstp);
}



/**
   copy a char array into a field, replacing the text handle of the field.

   @param str string to put in the field (not necessarily terminated by 0)
   @param length number of bytes to copy
   @param fieldid id of the field
@return nothing
 */

void
StrNToField (char *str, UInt16 length, UInt16 fieldid)
{
    FieldPtr fdPtr = (FieldPtr) (GetObjPtr (fieldid));
    MemHandle OldH,
      NewH;
    FieldAttrType fdattr;

    debug (GPSUDBG_CALLS, ("<->StrNToField(l=%d, fieldid=%d)", length, fieldid));
    OldH = /* (VoidHand) */ FldGetTextHandle (fdPtr);
    NewH = MemHandleNew (length + 2);
    if (!NewH)
	return;
    StrNCopySafe (MemHandleLock (NewH), str, length + 1);
    MemHandleUnlock (NewH);

    FldSetTextHandle (fdPtr, /* (Handle) */ NewH);
    FldGetAttributes (fdPtr, &fdattr);
    if (fdattr.visible)
	FldDrawField (fdPtr);
    if (OldH)
	MemHandleFree (OldH);
}							   /* StrNToField */


/**
   Draws a string at given place and width. If the string don't fit
   in given width an ellipsis is drawn (...) at the end.

@param string string to draw
@param  x left coordinate to start
@param  y top coordinate to start
@param  w max width to draw into
@param  RightAligned the string is right aligned if true, left aligned otherwise
@return nothing
 */

void
DrawCharsInWidth (char *string, Int16 x, Int16 y, UInt16 w, Boolean RightAligned)
{
    Int16 width = w,
      dotw,
      len = StrLen (string);
    Boolean fitWithinWidth;

    FntCharsInWidth (string, &width, &len, &fitWithinWidth);
    if (fitWithinWidth)
	WinDrawChars (string, len, RightAligned ? x + w - FntCharsWidth (string, len) : x, y);
    else
      {							   /* pas assez de place */
	  dotw = 3 * FntCharWidth ('.');
	  width = w - dotw;
	  FntCharsInWidth (string, &width, &len, &fitWithinWidth);
	  WinDrawChars (string, len, RightAligned ? x + w - width - dotw : x, y);
	  WinDrawChars ("...", 3, RightAligned ? x + w - dotw : x + width, y);
      }
}							   /* DrawCharsInWidth */


/**
   Canonization of a string (all upper case, no non-alphanum,...)

@param p string to canonicalise
@param  l length of string
@return nothing
 */

void
Canonicalize (char *p, UInt16 l)
{
    Int16 i;
    for (i = 0; i < l; i++, p++)
	if ((*p >= 'a') && (*p <= 'z'))
	    *p -= 'a' - 'A';				   /* conversion en majuscules */
	else if (((*p < 'A') || (*p > 'Z')) && (*p != '-') && (*p != ' ')
		 && ((*p < '0') || (*p > '9')))
	    *p = ' ';
}							   /* Canonicalize */


		      // ********* Forms utilities :

/**
   Get object form pointer from object ID

   @param  ObjID object id
   @return pointer on form object
 */

void *
GetObjPtr (UInt16 ObjID)
{
    FormPtr frm = FrmGetActiveForm ();
    debug (GPSUDBG_CALLS, ("<->GetObjPtr(%d)", ObjID));
    return (FrmGetObjectPtr (frm, (FrmGetObjectIndex (frm, ObjID))));
}


/**
   Hides given object in current form

   @param  objID object index to hide
   @return nothing
 */

void
HideObject (UInt16 objID)
{
    FormPtr frm = FrmGetActiveForm ();
    FrmHideObject (frm, FrmGetObjectIndex (frm, objID));
}


/**
   Show given object in current form

@param  objID object index to show
@return nothing
 */

void
ShowObject (UInt16 objID)
{
    FormPtr frm = FrmGetActiveForm ();
    FrmShowObject (frm, FrmGetObjectIndex (frm, objID));
}


/**
   Set a control's values

   @param  objID id of the control
   @param  value value to set
   @return nothing
*/

void
SetControlValue (UInt16 objID, short value)
{
    FormPtr frm = FrmGetActiveForm ();
    FrmSetControlValue (frm, FrmGetObjectIndex (frm, objID), value);
}


/**
   Get a control's (push button or check box) state

   @param  objID id of the control
   @return the state of the control
*/

short
GetControlValue (UInt16 objID)
{
    FormPtr frm = FrmGetActiveForm ();
    return (FrmGetControlValue (frm, FrmGetObjectIndex (frm, objID)));
}


/**
   Get an object's bounds

   @param  objID object's ID in the form
   @param  rp pointer on a RectangleType struct to store bounds into
   @return nothing
 */


void
GetObjectBounds (UInt16 objID, RectanglePtr rp)
{
    FormPtr frm = FrmGetActiveForm ();
    FrmGetObjectBounds (frm, FrmGetObjectIndex (frm, objID), rp);
}



		      // ********* record utilities :

/**
      Returns a record position in it's category and index 
  given it's ID (avoids DmFindRecordByID pitfalls)

   @param DB data base to search
   @param category category to search into
   @param recid id of record to search
   @param indexp pointer on index to return if found, unchanged otherwise. May be null
   @return position value if found (and *indexp updated), -1 otherwise
 */

Int16
RecIdToPosition (DmOpenRef DB, UInt16 category, UInt32 recid, UInt16 * indexp)
{
    UInt16 i = 0;
    UInt16 pos = 0;
    UInt32 l;
    UInt16 attr;

    debug (GPSUDBG_CALLS, ("RecIdToPosition(recid=%ld)", recid));
    while (DmQueryNextInCategory (DB, &i, category))
      {
	  DmRecordInfo (DB, i, &attr, &l, NULL);
	  if (l == recid)
	    {
		if (indexp)
		    *indexp = i;
		return (pos);
	    }
	  pos++;
	  i++;
      }
    return (-1);

}


/**
   Get unique ID of a record given it's index

   @param  DB reference to data base to search
   @param  index index of the record
   @return UID of the record or 0 if record not found and OS not crashed
 */

UInt32
GetRecordID (DmOpenRef DB, UInt16 index)
{
    UInt32 recid = 0;
    DmRecordInfo (DB, index, NULL, &recid, NULL);
    debug (GPSUDBG_CALLS, ("GetRecordID(%d)->%ld", index, recid));
    return (recid);
}



		      // ********* Floating point related functions :

/**
   Conversion of a double to a string (fixed precision) of the form [-]ddd.ffff

   @param  x value to convert (we suppose x < 1e9)
   @param  digits number of digits of integer part (ddd), negative sign included. Must be <=9.
           Exemple : x=-34, digits=3 => "-34." If digits is 0 the necessary number of digits
           is returned.
   @param  precision number of digits of fractional part (ffff). Must be <=9
   @return the string containing value converted
 */

char *
DToA (double x, UInt16 digits, UInt16 precision)
{
    UInt32 l,
      k;
    UInt16 i;
    Boolean sign = 1;
    static char buf[21];	       /* ['-' +] 9 digits + '.' + 9 digits + 0 */
    char *p,
     *first;

    debug_enter_func (GPSUDBG_CALLS, "(digits=%d,prec=%d)", digits, precision);
    MemSet (buf, sizeof (buf), '0');
    buf[9] = '.';
    if (x < 0.0)
	x = -x;
    else
	sign = 0;
    digits = min (digits, 9);
    precision = min (precision, 9);
    buf[9 + precision + 1] = 0;
    for (i = precision, k = 1; i; i--)
	k *= 10;
    x += (double) (1. / (2. * (double) k));		   /* arrondi */
    l = (long) x;					   /* partie entire */
    p = buf + 8;
    for (; l; l /= 10, p--)
	*p = (l % 10) + '0';
    first = min (p + 1, buf + 8);
    x -= (long) x;					   /* partie fractionnaire */
    l = x * (double) k;
    p = buf + 9 + precision;
    //  *(p+1)=0;
    for (; l; l /= 10, p--)
	*p = (l % 10) + '0';
    p = digits ? buf + 9 - digits : first;
    if (sign)
	*--p = '-';
    debug_exit_func (GPSUDBG_CALLS, "=\"%s\"", p);
    return (p);
}


/**
   Conversion ascii floating point number ([+-]M[M...][.F[F...]][+-]eX[X]) to  double

 @param p string to parse
 @param endp si non null address of a pointer to store where the conversion has stopped
 @return result of conversion
*/

double
StrToD (char *p, char **endp)
{
    double d = 0.;
    enum
    { get_mantisse, get_frac, get_exp }
    state = get_mantisse;
    Boolean empty = 1,
      m_sign = 0,
      e_sign = 0;
    Int16 i,
      stop = 0;
    UInt32 coeff = 10,
      ecoeff = 10;
    //    WordPtr Cattr = GetCharAttr ();

    debug_enter_func (GPSUDBG_CALLS, "(\"%s\"), test:%d", p,TxtGlueCharIsDigit (''));
    while (*p && TxtGlueCharIsSpace (*p))
	p++;						   // on saute les espaces eventuels

    for (; *p && !stop; p++)
      {
	  switch (state)
	    {
	    case get_mantisse:
		if (TxtGlueCharIsDigit (*p))
		  {
		      d = d * 10. + (*p - '0');
		      empty = 0;
		  }
		else if (*p == '.')
		    state = get_frac;
/*  	    else if (empty) */
/*  		stop = 1; */
		else if (*p == '-')
		    m_sign = 1;
		else if (*p == '+');
		else if (*p == 'e')
		    state = get_exp;
		else
		    stop = 1;
		break;
	    case get_frac:
		if (TxtGlueCharIsDigit (*p))
		    d = d + (double) (*p - '0') / coeff;
		else if (*p == 'e')
		    state = get_exp;
		else
		    stop = 1;
		coeff *= 10;
		break;
	    case get_exp:
		if (TxtGlueCharIsDigit (*p))
		  {
		      for (i = 0; i < (*p - '0'); i++)
			  d = e_sign ? d / ecoeff : d * ecoeff;
		      ecoeff = e_sign ? ecoeff / 10 : ecoeff * 10;
		  }
		else if (*p == '-')
		    e_sign = 1;
		else if (*p == '+');
		else
		    stop = 1;
	    }						   /* switch */
      }
    if (stop)
	p--;
    if (endp)
	*endp = p;
    debug_exit_func (GPSUDBG_CALLS, "(\"%s\")=%s",endp?*endp:"",DToA (m_sign?-d:d,0,3));
    return (m_sign?-d:d);
}							   // StrToD()


/**
   Converts ASCII coordinate to double. Any format (D.DD, DM.MM, DMS) is accepted.


   @param  string pointer on coordinates
   @param  longitude pointer on boolean set to true if it's a longitude coordinate, false otherwise. 
           If no indication in string, value passed is used.
   @return coordinate if conversion ok, BadCoordinate otherwise
 */

double
ReadCoord (char *string, Boolean * longitude)
{
    char *stop = string + StrLen (string);
    enum
    { deg = 1, min = 2, sec = 4, latlon = 8 }
    got = 0;
    double t,
      r = 0.,
      coef = 1;
    char *dep = string;

    debug (GPSUDBG_CALLS, ("<->ReadCoord(\"%s\",\"%s\")", string, longitude ? "longitude" : "latitude"));
    while (string < stop)
      {
	  dep = string;
	  t = StrToD (string, &string);

	  while (*string == ' ')
	      string++;
	  switch (*string)
	    {
	    case '':
		if (got & deg)
		    return (BadCoordinate);
		got |= deg;
		r += t;
		break;
	    case '\'':
		if (got & min)
		    return (BadCoordinate);
		got |= min;
		r += t / 60.;
		break;
	    case '"':
		if (got & sec)
		    return (BadCoordinate);
		got |= sec;
		r += t / 3600.;
		break;
	    case 'N':
	    case 'n':
		if (got & latlon)
		    return (BadCoordinate);
		got |= latlon;
		*longitude = 0;
		break;
	    case 'S':
	    case 's':
		if (got & latlon)
		    return (BadCoordinate);
		got |= latlon;
		*longitude = 0;
		coef = -1;
		break;
	    case 'W':
	    case 'w':
		if (got & latlon)
		    return (BadCoordinate);
		got |= latlon;
		*longitude = 1;
		coef = -1;
		break;
	    case 'E':
	    case 'e':
		if (got & latlon)
		    return (BadCoordinate);
		got |= latlon;
		*longitude = 1;
		break;
	    default:
		if (!(got & latlon) || !(got & 7))
		    return (BadCoordinate);
		else
		    stop = string;
	    }						   /* switch */
	  string++;
      }							   /* while */
    if (!*longitude && (r > 90.))
	return (BadCoordinate);
    if (*longitude && (r > 180.))
	return (BadCoordinate);
    return (r * coef);
}							   /* ReadCoord */


		      // ********* coordinates display utilities :

/**
   Converts coordinate to string in current format (specified in DisplayMode)

   @param  coord coordinate in degrees
   @param  longitude true indicates coordinate is a longitude
   @param  DisplayMode display format wanted
   @return converted string
 */


char *
DegreToChars (double coord, Boolean longitude, DisplayFormat DisplayMode)
{
    switch (DisplayMode)
      {
      case DMM:
	  return (DegreToDM (coord, longitude, 0));
	  break;
      case DMS:
	  return (DegreToDMS (coord, longitude));
	  break;
      case DDD:
      default:
	  return (DegreToD (coord, longitude));
      }
}							   /* DegreToChars */


/**
   Conversion of coordinate in degrees to ascii format [D]DDMM.fff'L
 or  [D]DDMM.fff,L if nmea

@param  coord coordinate in degrees
@param  longitude true if longitude coordinate
@param  nmea true if nmea format wanted
@return pointer on resulting string
*/

char *
DegreToDM (double coord, Boolean longitude, Boolean nmea)
{
    Int16 d,
      m,
      mm;
    static char buf[16];
    char c,
      deg[2],
      e;

    c = longitude ? coord < 0 ? 'W' : 'E' : coord < 0 ? 'S' : 'N';
    if (coord < 0)
	coord = -coord;
    d = (Int16) coord;					   /* degres */
    coord -= d;						   /* fractions de degres */
    coord *= 60.0;					   /* minutes */
    m = (Int16) coord;					   /* partie entiere de mn */
    coord -= m;						   /* fractions de minutes */
    coord *= 1000;					   /* milliemes de mn */
    mm = (Int16) (coord + 0.5);				   /* partie entieres milliemes de mn */
    if (mm >= 1000)
      {							   /* arrondi provoque debordement ? */
	  mm = 0;
	  if (++m == 60)
	    {
		m = 0;
		d++;					   /* si oui, repercussions sur mn et degres */
	    }
      }
    deg[0] = nmea ? 0 : '';
    deg[1] = 0;
    e = nmea ? ',' : '\'';
    StrPrintF (buf, "%s%d%s%s%d.%s%d%c%c",
	       longitude ? d > 99 ? "" : d > 9 ? "0" : "00" : d > 9 ? "" : "0",
	       d, deg, m > 9 ? "" : "0", m, mm > 99 ? "" : mm > 9 ? "0" : "00", mm, e, c);
    debug (GPSUDBG_CALLS, ("<->DegreToDM()=\"%s\"", buf));
    return (buf);
}							   /* DegreToDM */


/**
  degrees to string  [D]DDMM'SS.S"L

@param  coord coordinate in degrees
@param  longitude true indicates coord is a longitude
@return result string
 */

char *
DegreToDMS (double coord, Boolean longitude)
{
    Int16 d,
      m,
      s,
      ss;
    static char buf[16];
    char c;
    double precision = (1. / 3600.) / 10.;	/* precision : 1/10s */

    c = longitude ? coord < 0 ? 'W' : 'E' : coord < 0 ? 'S' : 'N';
    if (coord < 0)
	coord = -coord;
    coord += precision / 2.;				   /* arrondissement preventif */
    d = (Int16) coord;					   /* degres */
    m = (Int16) (fmod (coord, 1) * 60.);			   /* minutes */
    s = (Int16) (fmod (coord, 1. / 60.) * 3600.);		   /* secondes */
    ss = (Int16) (fmod (coord, 1. / 3600.) * 36000.);	   /* dixiemes de s */
    StrPrintF (buf, "%s%d%s%d'%s%d.%d\"%c",
	       longitude ? d > 99 ? "" : d > 9 ? "0" : "00" : d > 9 ? "" : "0",
	       d, m > 9 ? "" : "0", m, s > 9 ? "" : "0", s, ss, c);
    debug (GPSUDBG_CALLS, ("<->DegreToDMS()=\"%s\"", buf));
    return (buf);
}							   /* DegreToDMS */


/**
   degrees to string  [D]DD.DDDDDL

@param  coord coordinate in degrees
@param  longitude true indicates coord is a longitude
@return result string
 */

char *
DegreToD (double coord, Boolean longitude)
{
    Int16 d;
    UInt32 dd;
    static char buf[16];
    char conv[8],
      fill[] = "000000";
    char c;
    double precision = 0.00001;	       /* precision : 1/100000 */

    c = longitude ? coord < 0 ? 'W' : 'E' : coord < 0 ? 'S' : 'N';
    if (coord < 0)
	coord = -coord;
    coord += precision / 2.;				   /* arrondissement preventif */
    d = (Int16) coord;					   /* degres */
    dd = (UInt32) (fmod (coord, 1) * 100000.);		   /* 1/100000 */
    StrIToA (conv, dd);
    StrPrintF (fill + (5 - StrLen (conv)), "%s", conv);
    StrPrintF (buf, "%s%d.%s%c",
	       longitude ? d > 99 ? "" : d > 9 ? "0" : "00" : d > 9 ? "" : "0", d, fill, c);
    debug (GPSUDBG_CALLS, ("<->DegreToD()=\"%s\"", buf));
    return (buf);
}							   /* DegreToD */


/**
   conversion of degree coordinates to Semicircle_Type

@param  latitude latitude in degrees
@param  longitude longitude in degrees
@param  p pointer to store the result
@return nothing
*/

void
DegreToSemi (double latitude, double longitude, Semicircle_Type * p)
{
    const double d2s = ((long) 1 << 30) / 90.0;

    debug (GPSUDBG_CALLS, ("<->DegreToSemi()"));
    p->lat = latitude * d2s + 0.5;
    p->lon = longitude * d2s + 0.5;
}


/**
   Conversion of coordinates from Semicircle_Type to degree

@param  p address of source coordinate
@param latitude pointer to store latitude
@param longitude pointer to store longitude
@return nothing
*/

void
SemiToDegre (Semicircle_Type * p, double *latitude, double *longitude)
{
    const double s2d = 90.0 / ((long) 1 << 30);

    debug (GPSUDBG_CALLS, ("SemiToDegre()"));
    *latitude = p->lat * s2d;
    *longitude = p->lon * s2d;
}


/**
   Computes Semicircle_Type coordinates from distance and bearing
 from a given point

@param dist distance in nm
@param  bearing bearing in degrees
@param  p address of origin then result coordinates
@return nothing
*/

void
DistBearToSemi (double dist, UInt16 bearing, Semicircle_Type * p)
{
    double const a = PI / OnePower31;
    double lat,
      lon,
      lat1 = p->lat * a,
      lon1 = -p->lon * a,
      tc = bearing * PI / 180.,
      d = PI / (180. * 60.) * dist;

    debug (GPSUDBG_CALLS, ("DistBearToSemi(bearing=%d)", bearing));
    lat = asin (sin (lat1) * cos (d) + cos (lat1) * sin (d) * cos (tc));
    if (cos (lat) == 0)
	lon = lon1;					   // endpoint a pole
    else
	lon = fmod (lon1 - asin (sin (tc) * sin (d) / cos (lat)) + PI, 2. * PI) - PI;
    p->lat = round (lat / a);
    p->lon = -round (lon / a);
}							   /* DistBearToSemi */


/**
   Bearing of point 2 from point 1

@param  slat1 latitude of point 1
@param  slon1 longitude of point 1
@param  slat2 latitude of point 2
@param  slon2 longitude of point 2
@return bearing in degrees
 */

UInt16
Bearing (long slat1, long slon1, long slat2, long slon2)
{
    double const a = PI / OnePower31;
    double lat1 = slat1 * a,
      lon1 = -slon1 * a,
      lat2 = slat2 * a,
      lon2 = -slon2 * a,
      b;
    Int16 r;

    b = fmod (atan2 (sin (lon1 - lon2) * cos (lat2),
		     cos (lat1) * sin (lat2) - sin (lat1) * cos (lat2) * cos (lon1 - lon2)),
	      2 * PI);
    r = round (180. / PI * b);
    debug (GPSUDBG_CALLS, ("Bearing()=%d", r >= 0 ? r : r + 360));
    return (r >= 0 ? r : r + 360);
}							   /* Bearing */


/**
   Distance comparison of two points against a reference point

   @param lat1 latitude of first point
   @param lon1 longitude of first point
   @param lat2 latitude of second point
   @param lon2 longitude of second point
   @param reflat latitude of reference point
   @param reflon longitude of reference point
   @return -1, 0 or 1
 */

Int16
DistComp (long lat1, long lon1, long lat2, long lon2, long reflat, long reflon)
{
    double d1,
      d2;
    d1 = PDistance (reflat, reflon, lat1, lon1);
    d2 = PDistance (reflat, reflon, lat2, lon2);
    return (d1 > d2 ? 1 : d1 == d2 ? 0 : -1);
}


/**
   Computes distance between two points given their coordinates

@param  lat1 latitude WGS84 of point 1
@param  lon1 longitude WGS84 of point 1
@param  lat2 latitude WGS84 of point 2
@param  lon2 longitude WGS84 of point 2
@param  unit unit wanted 
@return distance in km, nm or mi
 */

double
Distance (long lat1, long lon1, long lat2, long lon2, DistanceUnit unit)
{
    double lat,
      lon,
      rlat,
      rlon,
      u,
      v,
      d;
    static double const a = PI / OnePower31,
      bnm = 3437.746771,	       /* (180*60)/PI*1.000 (nm) */

      bm = 6366.707019,		       /* (180*60)/PI*1.852 (km) */

      bs = 3956.098164;		       /* (180*60)/PI*1.852/1.60934 (mi) */

    debug (GPSUDBG_CALLS, ("<->Distance(%ld,%ld,%ld,%ld,%d)", lat1, lon1, lat2, lon2, unit));
    rlat = lat1 * a;					   /* conversion en radians (factorisable ?) */
    rlon = lon1 * a;
    lat = lat2 * a;
    lon = lon2 * a;
    u = sin ((rlat - lat) / 2.);
    v = sin ((rlon - lon) / 2.);
    d = 2. * asin (sqrt (u * u + cos (rlat) * cos (lat) * v * v));
    if (unit == metric)
	return (d * bm);
    else if (unit == nautical)
	return (d * bnm);
    else
	return (d * bs);
}							   /* Distance */


/**
   Pseudo distance between 2 points.
 This is approximate but fast !

@param  lat1 latitude of first point
@param  lon1 longitute of first point
@param  lat2 latitude of second point
@param  lon2 longitute of second point
@return pseudo-distance
*/

double
PDistance (long lat1, long lon1, long lat2, long lon2)
{
    double dx,
      dy;
    long l;
/* table de cosinus "simplifie" pour calcul de pseudo-distance : */
    static /*const */ double Pcos[33] =
    { 1.000000, 0.998795, 0.995185, 0.989176, 0.980785, 0.970031, 0.956940, 0.941544, 0.923879,
	0.903989, 0.881921, 0.857728, 0.831469, 0.803207, 0.773009, 0.740950, 0.707105,
	0.671558, 0.634392, 0.595698, 0.555568, 0.514101, 0.471395, 0.427553, 0.382681,
	0.336887, 0.290282, 0.242977, 0.195087, 0.146727, 0.098014, 0.049064, 0
    };

    debug (GPSUDBG_CALLS, ("PDistance()"));
    dx = lon1 - lon2;
    l = (lat1 + lat2) / 2;
    if (l < 0)
	l = -l;
    dx *= Pcos[l / ((long) 1 << 25)];			   /* correction grossiere en latitude */
    dy = lat1 - lat2;
    return (dx * dx + dy * dy);
}							   /* PDistance */


		      // ********* Debugging :

/**
   Sets the debugging level (no effect if compiled without DEBUG)

@param debuglevel level of debugging (see values near top of nokialib.c)
@param logprocp procedure to call to log traces
@return nothing
 */

void
GpsutilSetDebugLevel (UInt16 debuglevel, LogProcPtr logprocp)
{
#ifdef DEBUG
    char debuf[256];
    DebugLevel = debuglevel;
    LogOutput = logprocp;
    DebugIndentLevel = 0;
    StrPrintF (debuf, "%s, debugging level set to %d", rcsid, DebugLevel);
    DebugPrintf (debuf);
#endif
}							   /* GpsutilSetDebugLevel */



/**
   Internal use. Writes "printf" formatted messages to debug log


 @param  formatStr Pointer to the format specification string
 @param ... parameters
 @return nothing
 */

void
DebugPrintf (char * formatStr, ...)
{
#ifdef DEBUG
    va_list args;
    char buf[256];

    if (DebugIndentLevel < 0)
	DebugIndentLevel = 0;
    else if (DebugIndentLevel > 20)
	DebugIndentLevel = 20;
    MemSet (buf, DebugIndentLevel, ' ');
    va_start (args, formatStr);
    StrVPrintF (buf + DebugIndentLevel, formatStr, args);
    va_end (args);
    if (LogOutput)
	LogOutput (buf);
#endif
}							   /* DebugPrintf */


/**
   Test if position is inside waypoint's proximity distance

   @param wptp pointer on a Custom_Wpt_Type waypoint
   @param  position pointer on a position
   @return true if position is inside proximity circle, flase otherwise
 */

Boolean
WptHit (Custom_Wpt_Type * wptp, Semicircle_Type * position)
{

    if (!wptp->dst)
	return (0);
    if (Distance (wptp->posn.lat, wptp->posn.lon, position->lat, position->lon, metric) <=
	(wptp->dst / 1000.))
	return (1);
    else
	return (0);
}
