/** Objects management of GPilotS.
  @file objects.c
  @author C. Claveleira
  @date 1999-2004
  
  Contains all sorts of format conversion routines for waypoints,
  routes and tracks from/to native Garmin formats to/from GPilotS formats
  and size and offset utilities,...

*/

/* 
   Copyright (C) 1999-2004 Christian Claveleira (Christian.Claveleira@cru.fr)

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
   
 */

/* 	$Id: objects.c,v 1.9 2004/11/17 21:04:30 clavelei Exp $	 */

#ifndef lint
static char vcid[] = "$Id: objects.c,v 1.9 2004/11/17 21:04:30 clavelei Exp $";
#endif /* lint */

#if defined __palmos__ 
# include <PalmOS.h>
# define min(a, b) (((a) < (b)) ? (a) : (b))
#elif defined GLIB
# include <glib.h>
# include <string.h>
#endif

#include "Garmin.h"
#include "types.h"
#include "log.h"
#include "gpsutil.h"

static const UInt16 smbl2ST[16] =
    { sym_wpt_dot, sym_house, sym_fuel, sym_car, sym_fish, sym_wreck, sym_anchor, sym_wreck,
    sym_user_exit, sym_skull, sym_flag, sym_camp, sym_circle_x, sym_deer, sym_1st_aid,
    sym_trcbck
};				       /* mapping between Symbol_Type and smbl */

// mapping of symbols id between Garmin id and local bitmap index :
static const UInt16 SymbolMap[][3] = {
    {sym_anchor, sym_rbcn, 0},
    {sym_boat_ramp, sym_circle_x, 38},
    {sym_is_hwy, sym_border, 68},
    {sym_airport, sym_seaplane, 131}
};
#define S_BEGIN 0					   // begin of segment
#define S_END 1						   // end of segment
#define S_LOCAL 2					   // local bitmap index for begining of segment
#define S_DEFAULT_LID  sym_null				   // defaultbitmap index

#define HOSTtoLE32(to,from) LEtoHOST32((to),(from))
#define HOSTtoLE16(to,from) LEtoHOST16((to),(from))

static void RawD301ToCustomTrkPt (Raw_D301_Trk_Point_Type * from, Custom_Trk_Point_Type * to) UTIL_SECTION;
static void RawD301ToCompactTrkPt (Raw_D301_Trk_Point_Type * from, Compact_Trk_Point_Type * to) UTIL_SECTION;
static void RawD300ToCustomTrkPt (Raw_D300_Trk_Point_Type * from, Custom_Trk_Point_Type * to) UTIL_SECTION;
static void RawD300ToCompactTrkPt (Raw_D300_Trk_Point_Type * from, Compact_Trk_Point_Type * to) UTIL_SECTION;
static void LEtoHOST32 (byte * pto, byte * pfrom) UTIL_SECTION;
static void LEtoHOST16 (byte * pto, byte * pfrom) UTIL_SECTION;


			  // *********     GPilotS custom objects management :


/**
   Compute size of GPilotS' objects. The size is rounded modulo 4 to avoid alignment problems.

@param  p pointer on object
@return size of object pointed by p
 */

UInt16
SizeOfObject (ObjectPtr p)
{
    UInt16 s;
    debug_in (128, ("->SizeOfObject(%ld)", p));
    s = p->header.size;
    s += (4 - s % 4) & 0x3;				   /* pour obtenir size multiple de 4 (pb alignements) */
    s += sizeof (p->header);
    debug_out (128, ("<->SizeOfObjec (%ld)= %d", p, s));
    return (s);
}


#if defined __palmos__

/**
   Returns number of Object type objects in a record

   @param  RecH Handle on a data base record
   @return number of objects or 0
 */

UInt16
ObjectsInRec (MemHandle RecH)
{
    byte *p,
     *stop;
    Int16 i = 0;

    debug (32, ("ObjectsInRec(%ld)", RecH));
    if (!RecH)
	return (0);
    p = MemHandleLock (RecH);
    for (stop = p + MemHandleSize (RecH); p < stop; p += SizeOfObject ((ObjectPtr) p))
	i++;
    MemHandleUnlock (RecH);
    debug (32, ("<->ObjectsInRec (%ld)=%d", RecH, i));
    return (i);
}


/**
   Computes offset of nth object in a record


   @param  RecH Handle on record containing objects
   @param  n index of object (first =0) to know offset of
   @return offset in bytes from **RecH
 */

UInt16
OffsetOfObjectInRec (MemHandle RecH, UInt16 n)
{
    byte *p,
     *start,
     *stop;

    if (!RecH)
	return (0);
    debug_in (32, ("->OffsetOfObjectInRec(n=%d)", n));
    start = p = MemHandleLock (RecH);
    for (stop = p + MemHandleSize (RecH); n && (p < stop); p += SizeOfObject ((ObjectPtr) p))
	n--;
    MemHandleUnlock (RecH);
    debug_out (32, ("<-OffsetOfObjectInRec returns %d", p - start));
    return (p - start);
}

#endif


		       // *********  Waypoint objects management :


/**
   Converts a D100 waypoint to GPilotS' internal format

  @param  p100 pointer on a raw D100 waypoint (as received from gps) to convert
  @param  pcust pointer on a custom (internal) waypoint to receive result of conversion
  @return size of custom waypoint type
*/

UInt16
G100ToCustom (Raw_D100_Wpt_Type * p100, Custom_Wpt_Type * pcust)
{
    Int16 i;

    if (!p100 || !pcust)
	return (0);
    MemSet (pcust, sizeof (Custom_Wpt_Type), 0);
    for (i = sizeof (p100->ident) - 1; i >= 0; i--)
	if (p100->ident[i] != ' ')
	    break;					   /* to strip blanks */
    StrNCopySafe (pcust->ident, p100->ident, i + 2);	   /* name */
    LEtoHOST32 ((byte *) & (pcust->posn.lat), p100->lat);	   /* latitude */
    LEtoHOST32 ((byte *) & (pcust->posn.lon), p100->lon);	   /* longitude */
    StrNCopySafe (pcust->cmnt, p100->cmnt, sizeof (p100->cmnt));	/* comment */
    return (sizeof (Custom_Wpt_Type));
}							   /* G100ToCustom */


/**
   Converts a GPilotS' internal format waypoint to a D100 one

  @param  pcust pointer on the custom waypoint to convert
  @param  p100 pointer on a raw D100 waypoint to receive result of conversion
  @return size of converted waypoint
 */

UInt16
CustomToG100 (Custom_Wpt_Type * pcust, Raw_D100_Wpt_Type * p100)
{
    if (!p100 || !pcust)
	return (0);
    MemSet (p100, sizeof (Raw_D100_Wpt_Type), 0);
    MemSet (p100->ident, sizeof (p100->ident), ' ');	   /* name */
    MemMove (p100->ident, pcust->ident, StrLen (pcust->ident));
    HOSTtoLE32 (p100->lat, (byte *) & (pcust->posn.lat));	   /* latitude */
    HOSTtoLE32 (p100->lon, (byte *) & (pcust->posn.lon));	   /* longitude */
    MemSet (p100->cmnt, sizeof (p100->cmnt), ' ');	   /* comment */
    MemMove (p100->cmnt, pcust->cmnt, StrLen (pcust->cmnt));
    return (sizeof (Raw_D100_Wpt_Type));
}							   /* CustomToG100 */


/**
   Converts a D102 waypoint to GPilotS' internal format

  @param  p102 pointer on a raw D102 waypoint (as received from gps) to convert
  @param  pcust pointer on a custom (internal) waypoint to receive result of conversion
  @return size of custom waypoint type
*/

UInt16
G102ToCustom (Raw_D102_Wpt_Type * p102, Custom_Wpt_Type * pcust)
{
    if (!G100ToCustom ((Raw_D100_Wpt_Type *) p102, pcust))
	return (0);
    LEtoHOST32 ((byte *) & (pcust->dst), p102->dst);		   /* distance */
    LEtoHOST16 ((byte *) & (pcust->smbl), p102->smbl);	   /* symbol id */
    return (sizeof (Custom_Wpt_Type));
}


/**
   Converts a GPilotS' internal format waypoint to a D102 one

  @param  pcust pointer on the custom waypoint to convert
  @param  p102 pointer on a raw D102 waypoint to receive result of conversion
  @return size of converted waypoint
 */

UInt16
CustomToG102 (Custom_Wpt_Type * pcust, Raw_D102_Wpt_Type * p102)
{
    if (!CustomToG100 (pcust, (Raw_D100_Wpt_Type *) p102))
	return (0);
    HOSTtoLE32 (p102->dst, (byte *) & (pcust->dst));		   /* distance */
    HOSTtoLE16 (p102->smbl, (byte *) & (pcust->smbl));	   /* symbol id */
    return (sizeof (Raw_D102_Wpt_Type));
}


/**
   Converts a D103 waypoint to GPilotS' internal format

  @param  p103 pointer on a raw D103 waypoint (as received from gps) to convert
  @param  pcust pointer on a custom (internal) waypoint to receive result of conversion
  @return size of custom waypoint type
*/

UInt16
G103ToCustom (Raw_D103_Wpt_Type * p103, Custom_Wpt_Type * pcust)
{

    if (!G100ToCustom ((Raw_D100_Wpt_Type *) p103, pcust))
	return (0);
    pcust->smbl = smblToSymbolType (p103->smbl);	   /* symbol id */
    pcust->dspl = p103->dspl;				   /* display option */
    return (sizeof (Custom_Wpt_Type));
}


/**
   Converts a GPilotS' internal format waypoint to a D103 one

  @param  pcust pointer on the custom waypoint to convert
  @param  p103 pointer on a raw D103 waypoint to receive result of conversion
  @return size of converted waypoint
*/

UInt16
CustomToG103 (Custom_Wpt_Type * pcust, Raw_D103_Wpt_Type * p103)
{
    if (!CustomToG100 (pcust, (Raw_D100_Wpt_Type *) p103))
	return (0);
    p103->smbl = SymbolTypeTosmbl (pcust->smbl);	   /* symbol id */
    p103->dspl = pcust->dspl;				   /* display option */
    return (sizeof (Raw_D103_Wpt_Type));
}


/**
   Converts a D104 waypoint to GPilotS' internal format

  @param  p104 pointer on a raw D104 waypoint (as received from gps) to convert
  @param  pcust pointer on a custom (internal) waypoint to receive result of conversion
  @return size of custom waypoint type
*/

UInt16
G104ToCustom (Raw_D104_Wpt_Type * p104, Custom_Wpt_Type * pcust)
{
    if (!G102ToCustom ((Raw_D102_Wpt_Type *) p104, pcust))
	return (0);
    pcust->dspl = p104->dspl;				   /* display option */
    return (sizeof (Custom_Wpt_Type));
}


/**
   Converts a GPilotS' internal format waypoint to a D104 one

  @param  pcust pointer on the custom waypoint to convert
  @param  p104 pointer on a raw D104 waypoint to receive result of conversion
  @return size of converted waypoint
*/

UInt16
CustomToG104 (Custom_Wpt_Type * pcust, Raw_D104_Wpt_Type * p104)
{
    if (!CustomToG102 (pcust, (Raw_D102_Wpt_Type *) p104))
	return (0);
    p104->dspl = pcust->dspl;				   /* display option */
    return (sizeof (Raw_D104_Wpt_Type));
}							   /* CustomToG104 */


/**
   Converts a D105 waypoint to GPilotS' internal format

  @param  p105 pointer on a raw D105 waypoint (as received from gps) to convert
  @param  pcust pointer on a custom (internal) waypoint to receive result of conversion
  @return size of custom waypoint type
*/

UInt16
G105ToCustom (Raw_D105_Wpt_Type * p105, Custom_Wpt_Type * pcust)
{
    if (!p105 || !pcust)
	return (0);
    MemSet (pcust, sizeof (Custom_Wpt_Type), 0);
    LEtoHOST32 ((byte *) & (pcust->posn.lat), p105->lat);	   /* latitude */
    LEtoHOST32 ((byte *) & (pcust->posn.lon), p105->lon);	   /* longitude */
    LEtoHOST16 ((byte *) & (pcust->smbl), p105->smbl);	   /* symbol id */
    StrNCopySafe (pcust->ident, p105->wpt_ident, sizeof (pcust->ident));	/* name */
    return (sizeof (Custom_Wpt_Type));
}							   /* G105ToCustom */


/**
   Converts a GPilotS' internal format waypoint to a D105 one

  @param  pcust pointer on the custom waypoint to convert
  @param  p105 pointer on a raw D105 waypoint to receive result of conversion
  @return size of converted waypoint
*/

UInt16
CustomToG105 (Custom_Wpt_Type * pcust, Raw_D105_Wpt_Type * p105)
{
    char *p;

    if (!p105 || !pcust)
	return (0);
    MemSet (p105, sizeof (Raw_D105_Wpt_Type), 0);
    p = StrNCopySafe (p105->wpt_ident, pcust->ident, sizeof (p105->wpt_ident));	/* ident */
    HOSTtoLE32 (p105->lat, (byte *) & (pcust->posn.lat));	   /* latitude */
    HOSTtoLE32 (p105->lon, (byte *) & (pcust->posn.lon));	   /* longitude */
    HOSTtoLE16 (p105->smbl, (byte *) & (pcust->smbl));	   /* symbol id */
    return (p - (char *) p105);
}							   /* CustomToG105 */


/**
   Converts a D107 waypoint to GPilotS' internal format

  @param  p107 pointer on a raw D107 waypoint (as received from gps) to convert
  @param  pcust pointer on a custom (internal) waypoint to receive result of conversion
  @return size of custom waypoint type
*/

UInt16
G107ToCustom (Raw_D107_Wpt_Type * p107, Custom_Wpt_Type * pcust)
{
    if (!G103ToCustom ((Raw_D103_Wpt_Type *) p107, pcust))
	return (0);
    LEtoHOST32 ((byte *) & (pcust->dst), p107->dst);		   /* distance */
    pcust->color = p107->color;				   /* color */
    return (sizeof (Custom_Wpt_Type));
}							   /* G107ToCustom */


/**
   Converts a GPilotS' internal format waypoint to a D107 one

  @param  pcust pointer on the custom waypoint to convert
  @param  p107 pointer on a raw D107 waypoint to receive result of conversion
  @return size of converted waypoint
*/

UInt16
CustomToG107 (Custom_Wpt_Type * pcust, Raw_D107_Wpt_Type * p107)
{
    if (!CustomToG103 (pcust, (Raw_D103_Wpt_Type *) p107))
	return (0);
    HOSTtoLE32 (p107->dst, (byte *) & (pcust->dst));		   /* distance */
    p107->color = pcust->color;				   /* color */
    return (sizeof (Raw_D107_Wpt_Type));
}							   /* CustomToG107 */


/**
   Converts a D108 waypoint to GPilotS' internal format

  @param  p108 pointer on a raw D108 waypoint (as received from gps) to convert
  @param  pcust pointer on a custom (internal) waypoint to receive result of conversion
  @return size of custom waypoint type
*/

UInt16
G108ToCustom (Raw_D108_Wpt_Type * p108, Custom_Wpt_Type * pcust)
{
    char *namep,
     *commentp;

    if (!p108 || !pcust)
	return (0);
    MemSet (pcust, sizeof (Custom_Wpt_Type), 0);
    //    namep = (char *) (&(p108->cc) + sizeof (p108->cc));
    namep = (char *) (p108 + 1);
    if (StrLen (namep) < 51)
      {
	  StrNCopySafe (pcust->ident, namep, sizeof (pcust->ident));	/* name */
	  commentp = namep + StrLen (namep) + 1;
	  StrNCopySafe (pcust->cmnt, commentp, sizeof (pcust->cmnt));	/* comment */
      }
    LEtoHOST32 ((byte *) & (pcust->posn.lat), p108->lat);	   /* latitude */
    LEtoHOST32 ((byte *) & (pcust->posn.lon), p108->lon);	   /* longitude */
    LEtoHOST32 ((byte *) & (pcust->dst), p108->dist);	   /* distance */
    LEtoHOST32 ((byte *) & (pcust->alt), p108->alt);		   /* altitude */
    LEtoHOST16 ((byte *) & (pcust->smbl), p108->smbl);	   /* symbol id */
    pcust->dspl = p108->dspl;				   /* display option */
    pcust->color = p108->color;				   /* color */
    return (sizeof (Custom_Wpt_Type));
}							   /* G108ToCustom */


/**
   Converts, and optionaly merge, a Custom_Wpt_Type waypoint to a B108_Wpt_Type one
   B108_Wpt_Type has the necessary size to hold a "full" Raw_D108_Wpt_Type

  @param  pcust pointer on the Custom_Wpt_Type waypoint to convert
  @param  p108 pointer on the destination B108_Wpt_Type waypoint (may contain a valid Raw_D108_Wpt_Type waypoint)
  @param  merge if != 0 the source waypoint is merged to the Raw_D108_Wpt_Type waypoint ccntained in the destination waypoint
  @return the size of the corresponding Raw_D108_Wpt_Type
*/

UInt16
CustomToB108 (Custom_Wpt_Type * pcust, B108_Wpt_Type * p108, Boolean merge)
{
    static byte default_subclass[18] =
	{ 0, 0, 0, 0, 0, 0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF
    };
    char facility[32],
      city[26],
      addr[52],
      cross_road[52],
     *p;

    if (!p108 || !pcust)
	return (0);

    if (merge)
      {
	  p = p108->ident;				   /* ident */
	  p += StrLen (p);				   /* comment */
	  p += StrLen (p);				   /* facility */
	  if (p < ((char *) p108 + sizeof (B108_Wpt_Type)))
	      StrNCopySafe (facility, p, sizeof (facility));
	  p += StrLen (p);				   /* city */
	  if (p < ((char *) p108 + sizeof (B108_Wpt_Type)))
	      StrNCopySafe (city, p, sizeof (city));
	  p += StrLen (p);				   /* address */
	  if (p < ((char *) p108 + sizeof (B108_Wpt_Type)))
	      StrNCopySafe (addr, p, sizeof (addr));
	  p += StrLen (p);				   /* cross_road */
	  if (p < ((char *) p108 + sizeof (B108_Wpt_Type)))
	      StrNCopySafe (cross_road, p, sizeof (cross_road));
	  p = StrNCopySafe (p108->ident, pcust->ident, sizeof (p108->ident));	/* ident */
	  p = StrNCopySafe (p, pcust->cmnt, sizeof (p108->comment));	/* comment */
	  p = StrNCopySafe (p, facility, sizeof (p108->facility));	/* facility */
	  p = StrNCopySafe (p, city, sizeof (p108->city)); /* city */
	  p = StrNCopySafe (p, addr, sizeof (p108->addr)); /* address */
	  p = StrNCopySafe (p, cross_road, sizeof (p108->cross_road));	/* cross_road */
      }

    else
      {
	  MemSet (p108, sizeof (B108_Wpt_Type), 0);
	  p108->w.attr = 0x60;
	  (void) MemMove (p108->w.subclass, &default_subclass, sizeof (p108->w.subclass));	/* default subclass */
	  p = StrNCopySafe (p108->ident, pcust->ident, sizeof (p108->ident));	/* ident */
	  p = StrNCopySafe (p, pcust->cmnt, sizeof (p108->comment));	/* comment */
	  p += 4;					   /* 4 empty strings */
      }
    p108->w.color = pcust->color;			   /* color */
    p108->w.dspl = pcust->dspl;				   /* display option */
    HOSTtoLE16 (p108->w.smbl, (byte *) & (pcust->smbl));	   /* symbol id */
    HOSTtoLE32 (p108->w.lat, (byte *) & (pcust->posn.lat));	   /* latitude */
    HOSTtoLE32 (p108->w.lon, (byte *) & (pcust->posn.lon));	   /* longitude */
    HOSTtoLE32 (p108->w.dist, (byte *) & (pcust->dst));	   /* distance */
    HOSTtoLE32 (p108->w.alt, (byte *) & (pcust->alt));	   /* altitude */
    return (p - (char *) p108);
}							   /* CustomToB108 */


/**
   Converts a D109 waypoint to GPilotS' internal format

  @param  p109 pointer on a raw D109 waypoint (as received from gps) to convert
  @param  pcust pointer on a custom (internal) waypoint to receive result of conversion
  @return size of custom waypoint type
*/

UInt16
G109ToCustom (Raw_D109_Wpt_Type * p109, Custom_Wpt_Type * pcust)
{
    char *namep,
     *commentp;

    if (!p109 || !pcust)
	return (0);
    MemSet (pcust, sizeof (Custom_Wpt_Type), 0);
    namep = (char *) (p109 + 1);
    if (StrLen (namep) < 51)
      {
	  StrNCopySafe (pcust->ident, namep, sizeof (pcust->ident));	/* name */
	  commentp = namep + StrLen (namep) + 1;
	  StrNCopySafe (pcust->cmnt, commentp, sizeof (pcust->cmnt));	/* comment */
      }
    LEtoHOST32 ((byte *) & (pcust->posn.lat), p109->lat);	   /* latitude */
    LEtoHOST32 ((byte *) & (pcust->posn.lon), p109->lon);	   /* longitude */
    LEtoHOST32 ((byte *) & (pcust->dst), p109->dist);	   /* distance */
    LEtoHOST32 ((byte *) & (pcust->alt), p109->alt);		   /* altitude */
    LEtoHOST16 ((byte *) & (pcust->smbl), p109->smbl);	   /* symbol id */
    pcust->dspl = (p109->dspl_color & 0x60) >> 5;	   /* display option */
    pcust->color = p109->dspl_color & 0xf;		   /* color */
    return (sizeof (Custom_Wpt_Type));
}							   /* G109ToCustom */


/**
   Converts, and optionaly merge, a Custom_Wpt_Type waypoint to a B109_Wpt_Type one
   B109_Wpt_Type has the necessary size to hold a "full" Raw_D109_Wpt_Type

  @param  pcust pointer on the Custom_Wpt_Type waypoint to convert
  @param  p109 pointer on the destination B109_Wpt_Type waypoint (may contain a valid Raw_D109_Wpt_Type waypoint)
  @param  merge if != 0 the source waypoint is merged to the Raw_D109_Wpt_Type waypoint ccntained in the destination waypoint
  @return the size of the corresponding Raw_D109_Wpt_Type
*/

UInt16
CustomToB109 (Custom_Wpt_Type * pcust, B109_Wpt_Type * p109, Boolean merge)
{
    static byte default_subclass[18] =
	{ 0, 0, 0, 0, 0, 0, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,
	0xFF
    };
    char facility[32],
      city[26],
      addr[52],
      cross_road[52],
     *p;

    if (!p109 || !pcust)
	return (0);

    if (merge)
      {
	  p = p109->ident;				   /* ident */
	  p += StrLen (p);				   /* comment */
	  p += StrLen (p);				   /* facility */
	  if (p < ((char *) p109 + sizeof (B109_Wpt_Type)))
	      StrNCopySafe (facility, p, sizeof (facility));
	  p += StrLen (p);				   /* city */
	  if (p < ((char *) p109 + sizeof (B109_Wpt_Type)))
	      StrNCopySafe (city, p, sizeof (city));
	  p += StrLen (p);				   /* address */
	  if (p < ((char *) p109 + sizeof (B109_Wpt_Type)))
	      StrNCopySafe (addr, p, sizeof (addr));
	  p += StrLen (p);				   /* cross_road */
	  if (p < ((char *) p109 + sizeof (B109_Wpt_Type)))
	      StrNCopySafe (cross_road, p, sizeof (cross_road));
	  p = StrNCopySafe (p109->ident, pcust->ident, sizeof (p109->ident));	/* ident */
	  p = StrNCopySafe (p, pcust->cmnt, sizeof (p109->comment));	/* comment */
	  p = StrNCopySafe (p, facility, sizeof (p109->facility));	/* facility */
	  p = StrNCopySafe (p, city, sizeof (p109->city)); /* city */
	  p = StrNCopySafe (p, addr, sizeof (p109->addr)); /* address */
	  p = StrNCopySafe (p, cross_road, sizeof (p109->cross_road));	/* cross_road */
      }

    else
      {
	  MemSet (p109, sizeof (B109_Wpt_Type), 0);
	  p109->w.dtyp = 0x01;				   /* data packet type */
	  p109->w.attr = 0x70;				   /* attributes */
	  (void) MemMove (p109->w.subclass, &default_subclass, sizeof (p109->w.subclass));	/* default subclass */
	  MemSet (p109->w.ete, sizeof (p109->w.ete), 0xff);	/* outbound link ete */
	  p = StrNCopySafe (p109->ident, pcust->ident, sizeof (p109->ident));	/* ident */
	  p = StrNCopySafe (p, pcust->cmnt, sizeof (p109->comment));	/* comment */
	  p += 4;					   /* 4 empty strings */
      }
    p109->w.dspl_color = pcust->color;			   /* color */
    p109->w.dspl_color |= (pcust->dspl << 5);		   /* display option */
    HOSTtoLE16 (p109->w.smbl, (byte *) & (pcust->smbl));	   /* symbol id */
    HOSTtoLE32 (p109->w.lat, (byte *) & (pcust->posn.lat));	   /* latitude */
    HOSTtoLE32 (p109->w.lon, (byte *) & (pcust->posn.lon));	   /* longitude */
    HOSTtoLE32 (p109->w.dist, (byte *) & (pcust->dst));	   /* distance */
    HOSTtoLE32 (p109->w.alt, (byte *) & (pcust->alt));	   /* altitude */
    return (p - (char *) p109);
}							   /* CustomToB109 */


/**
   Converts a waypoint object record, given a pointer on it, to a Custom_Wpt_Type type waypoint

   @param  recp pointeur on a waypoint record 
   @param  pcust pointer on a Custom_Wpt_Type waypoint
   @return the local format id of the record waypoint
 */

LocalFormatID
PtrToCustomWpt (ObjectPtr recp, Custom_Wpt_Type * pcust)
{
    debug (32, ("PtrToCustomWpt()"));
    if (!recp)
	return (Unknown);
    switch (recp->header.type)
      {
      case G100Type:
	  G100ToCustom (&(recp->d.g.d100), pcust);
	  return (G100Type);

      case G102Type:
	  G102ToCustom (&(recp->d.g.d102), pcust);
	  return (G102Type);

      case G103Type:
	  G103ToCustom (&(recp->d.g.d103), pcust);
	  return (G103Type);

      case G104Type:
	  G104ToCustom (&(recp->d.g.d104), pcust);
	  return (G104Type);

      case G105Type:
	  G105ToCustom (&(recp->d.g.d105), pcust);
	  return (G105Type);

      case G107Type:
	  G107ToCustom (&(recp->d.g.d107), pcust);
	  return (G107Type);

      case G108Type:
	  G108ToCustom (&(recp->d.g.d108), pcust);
	  return (G108Type);

      case G109Type:
	  G109ToCustom (&(recp->d.g.d109), pcust);
	  return (G109Type);

      default:
	  return (Unknown);
      }
}


/**
   Converts a waypoint object record, given a handle on it, to a Custom_Wpt_Type type waypoint

   @param  RecH Handle to a waypoint record
   @param  pcust pointer on a Custom_Wpt_Type waypoint
   @return the local format id of the original record waypoint
 */

LocalFormatID
HandleToCustomWpt (ObjectHandle RecH, Custom_Wpt_Type * pcust)
{
    LocalFormatID f = Unknown;

    debug (32, ("HandleToCustomWpt()"));
    if (!RecH)
	MemSet (pcust, sizeof (Custom_Wpt_Type), 0);
    else
      {
	  ObjectPtr RecP = MemHandleLock ((MemHandle) RecH);
	  f = PtrToCustomWpt (RecP, pcust);
	  MemHandleUnlock ((MemHandle) RecH);
      }
    return (f);
}							   /* HandleToCustomWpt */

#if defined __palmos__

/**
   Converts a waypoint object record, given its index in waypoint data base, to a Custom_Wpt_Type

   @param  index index of waypoint
   @param  DB pointer on data base containing the waypoint
   @param  pcust pointer on a Custom_Wpt_Type waypoint
   @return the local format id of the original record waypoint
 */

LocalFormatID
IndexToCustomWpt (UInt16 index, DmOpenRef DB, Custom_Wpt_Type * pcust)
{
    return (HandleToCustomWpt ((ObjectHandle) DmQueryRecord (DB, index), pcust));
}

#endif

/**
   Converts a Custom_Wpt_Type waypoint to an object of given format

   @param  pcust pointeur on Custom_Wpt_Type waypoint type
   @param  format format to convert to
   @param  newp pointeur on object to receive converted waypoint
   @param  merge If != 0 the Custom_Wpt_Type waypoint is to be merged with the content of *newp
   @return 0 if OK
 */

Err
CustomToAnyWpt (Custom_Wpt_Type * pcust, LocalFormatID format, ObjectPtr newp, Boolean merge)
{
    //    Err error;
    //    UInt16 size;
    //    LocalFormatID type;

    debug (32, ("CustomToAnyWpt(format=%d, merge=%s)", format, merge ? "1" : "0"));
    newp->header.type = format;

    switch (format)
      {
      case G100Type:
	  newp->header.size = CustomToG100 (pcust, &(newp->d.g.d100));
	  break;

      case G102Type:
	  newp->header.size = CustomToG102 (pcust, &(newp->d.g.d102));
	  break;

      case G104Type:
	  newp->header.size = CustomToG104 (pcust, &(newp->d.g.d104));
	  break;

      case G105Type:
	  newp->header.size = CustomToG105 (pcust, &(newp->d.g.d105));
	  break;

      case G107Type:
	  newp->header.size = CustomToG107 (pcust, &(newp->d.g.d107));
	  break;

      case G108Type:
	  newp->header.size = CustomToB108 (pcust, &(newp->d.b108), merge);
	  break;

      case G109Type:
	  newp->header.size = CustomToB109 (pcust, &(newp->d.b109), merge);
	  break;

      default:
	  newp->header.size = CustomToG103 (pcust, &(newp->d.g.d103));
      }							   /* switch (format) */

    return (0);
}							   /* CustomToAnyWpt */


/**
   Converts a waypoint contained in an object from any known format to any other format

   @param  from pointer on source object
   @param  to pointer on destination object
   @param  toformat format to convert to
   @return pointer on destination object or NULL if unsupported format
 */

ObjectPtr
AnyToAnyWpt (ObjectPtr from, ObjectPtr to, LocalFormatID toformat)
{
    Custom_Wpt_Type cust;

    debug (32, ("AnyToAnyWpt(toformat=%d)", toformat));
    if (from->header.type == toformat)
      {
	  MemMove (to, from, SizeOfObject (from));
	  return (to);
      }
    to->header.type = toformat;
    PtrToCustomWpt (from, &cust);
    switch (toformat)					   // factorisable $$
      {
      case G100Type:
	  to->header.size = CustomToG100 (&cust, &(to->d.g.d100));
	  break;

      case G102Type:
	  to->header.size = CustomToG102 (&cust, &(to->d.g.d102));
	  break;

      case G103Type:
	  to->header.size = CustomToG103 (&cust, &(to->d.g.d103));
	  break;

      case G104Type:
	  to->header.size = CustomToG104 (&cust, &(to->d.g.d104));
	  break;

      case G105Type:
	  to->header.size = CustomToG105 (&cust, &(to->d.g.d105));
	  break;

      case G107Type:
	  to->header.size = CustomToG107 (&cust, &(to->d.g.d107));
	  break;

      case G108Type:
	  to->header.size = CustomToB108 (&cust, &(to->d.b108), 0);
	  break;

      case G109Type:
	  to->header.size = CustomToB109 (&cust, &(to->d.b109), 0);
	  break;

      default:
	  to=NULL;
      }							   /* switch (format) */
    return (to);
}							   /* AnyToAnyWpt */


/**
   Maps a garmin Symbol_Type value into a local bitmap index

   @param  gid Garmin symbol value
   @return local bitmap index
 */

UInt16
symb_gid2lid (Symbol_Type gid)
{
    Int16 i;

    for (i = 0; i < 4; i++)
	if ((gid >= SymbolMap[i][S_BEGIN]) && (gid <= SymbolMap[i][S_END]))
	    return (SymbolMap[i][S_LOCAL] + (gid - SymbolMap[i][S_BEGIN]));
    return (S_DEFAULT_LID);
}							   /* symb_gid2lid */


/**
   Maps a local bitmap index to a garmin Symbol_Type value

   @param lid local bitmap index
   @return Garmin symbol value
 */

Symbol_Type
symb_lid2gid (UInt16 lid)
{
    Int16 i;
    for (i = 0; i < 4; i++)
	if (lid <= (SymbolMap[i][S_LOCAL] + SymbolMap[i][S_END] - SymbolMap[i][S_BEGIN]))
	    return (SymbolMap[i][S_BEGIN] + (lid - SymbolMap[i][S_LOCAL]));
    return (S_DEFAULT_LID);
}							   /* symb_lid2gid */



		       // *********  Routes objects management :

#if defined __palmos__

/**
   Computes number of waypoints in route record. The route is supposed to contain only waypoints.

   @param  RecH Handle on route record
   @return number of waypoints
 */

UInt16
WptsInRteDB (MemHandle RecH)
{
    UInt16 i;
    debug_in (32, ("->WptsInRteDB(%ld)", RecH));
    i = ObjectsInRec (RecH);
    debug_out (32, ("<-WptsInRteDB returns %d", i > 0 ? i - 1 : 0));
    return (i > 0 ? i - 1 : 0);				   /* # of waypoints = # of objects - 1 (header) */
}

#endif

/**
   Conversion of a route header object, given a pointer on it, to a Custom_Rte_Hdr_Type GPilotS' internal format

  @param recp pointer on the object containing the route header to convert
  @param  pcust pointer on a Custom_Rte_Hdr_Type
  @return the local format id of the initial route header
 */

LocalFormatID
PtrToCustomRteHdr (ObjectPtr recp, Custom_Rte_Hdr_Type * pcust)
{
    debug (32, ("PtrToCustomRteHdr()"));
    if (!recp)
	return (Unknown);
    MemSet (pcust, sizeof (Custom_Rte_Hdr_Type), 0);
    switch (recp->header.type)
      {
      case G201Type:
	  pcust->nmbr = recp->d.g.d201.nmbr;
	  StrNCopySafe (pcust->cmnt, recp->d.g.d201.cmnt, sizeof (recp->d.g.d201.cmnt));
	  return (G201Type);

      case G202Type:
	  pcust->nmbr = 1;
	  StrNCopySafe (pcust->cmnt, recp->d.g.d202.rte_ident, sizeof (pcust->cmnt));
	  return (G202Type);

      case CustomRteHdr:
	  MemMove (pcust, &recp->d.CustRteHdr, sizeof (Custom_Rte_Hdr_Type));
	  return (CustomRteHdr);

      default:
	  return (Unknown);
      }
}							   /* PtrToCustomRteHdr */


/**
   Converts route header object, given a handle on it, to Custom_Rte_Hdr_Type GPilotS' internal format

@param  RecH handle on route header object
@param pcust pointer on a Custom_Rte_Hdr_Type structure
@return format of origin object
 */

LocalFormatID
HandleToCustomRteHdr (MemHandle RecH, Custom_Rte_Hdr_Type * pcust)
{
    LocalFormatID f;
    MemPtr RecP;

    debug (32, ("HandleToCustomRteHdr()"));
    if (!RecH)
	return (Unknown);
    RecP = MemHandleLock (RecH);
    f = PtrToCustomRteHdr (RecP, pcust);
    MemHandleUnlock (RecH);
    return (f);
}							   /* HandleToCustomRteHdr */


/**
   Converts a custom GPilotS internal format route header to a known Garmin type included in an object

   @param pcust pointer on a Custom_Rte_Hdr_Type structure to convert
   @param  o pointer on a GPilotS' object to receive result of conversion
   @param  toformat format to convert to
   @return pointer on destination object
 */

ObjectPtr
CustomToRteHdrObj (Custom_Rte_Hdr_Type * pcust, ObjectPtr o, LocalFormatID toformat)
{
    debug (32, ("CustomToRteHdrObj()"));
    if (!pcust || !o)
	return (o);

    switch (toformat)
      {
      case G202Type:
	  o->header.type = G202Type;
	  StrNCopySafe (o->d.g.d202.rte_ident, pcust->cmnt, sizeof (o->d.g.d202.rte_ident));
	  o->header.size = StrLen (o->d.g.d202.rte_ident);
	  break;

      case G201Type:
      default:
	  o->header.type = G201Type;
	  o->header.size = sizeof (D201_Rte_Hdr_Type);
	  o->d.g.d201.nmbr = pcust->nmbr;
	  MemSet (o->d.g.d201.cmnt, sizeof (o->d.g.d201.cmnt), ' ');
	  MemMove (o->d.g.d201.cmnt, pcust->cmnt,
		   min (StrLen (pcust->cmnt), sizeof (o->d.g.d201.cmnt)));
	  break;
      }
    return (o);
}							   /* CustomToRteHdrObj */


/**
   Converts object containing a route header from any known format to any other

   @param  from pointer on source object containing a route header
   @param  to pointer on destination object to receive result of conversion
   @param  toformat format to convert to
   @return pointer on destination object
 */

ObjectPtr
AnyToAnyRteHdr (ObjectPtr from, ObjectPtr to, LocalFormatID toformat)
{
    Custom_Rte_Hdr_Type cust;
    //  ObjectPtr o;

    debug (32, ("AnyToAnyRteHdr()"));
    if (from->header.type == toformat)
	MemMove (to, from, SizeOfObject (from));
    else
      {
	  PtrToCustomRteHdr (from, &cust);
	  CustomToRteHdrObj (&cust, to, toformat);
      }
    return (to);
}							   /* AnyToAnyRteHdr */




		      // *********  Track objects management :

#if defined __palmos__

/**
   Returns number of track points in a track
 (should use number stored in track header)
   @param  rech handle on a record containing a track
  @return number of track points in the trace
 */

UInt16
TrkptsInTrkDB (MemHandle rech)
{							   /* nombre de trackpoints dans une trace */
    ObjectPtr RecP;
    UInt32 size = 0;
    UInt16 n = 0;

    if (!rech)
	return (0);
    RecP = MemHandleLock (rech);
    if (RecP->d.CustTrkHdr.type == CustomTrkPt)
	size = sizeof (Custom_Trk_Point_Type);
    else if (RecP->d.CustTrkHdr.type == CompactTrkPt)
	size = sizeof (Compact_Trk_Point_Type);
    if (size)
	n = (MemHandleSize (rech) - SizeOfObject (RecP)) / size;
    //  return ((MemHandleSize (rech) - sizeof (Custom_Trk_Hdr_Type)-sizeof(((ObjectPtr)0)->header)) / sizeof (Custom_Trk_Point_Type))    ;
    MemHandleUnlock (rech);
    debug (32, ("TrkptsInTrkDB(%ld)->%d", rech, n));
    return (n);
}               /* TrkptsInTrkDB */

#endif

/**
   Converts a raw D301 track point to GPilotS standard custom internal format

  @param  from pointer on raw D301 track point to convert
  @param  to pointer on a Custom_Trk_Point_Type struct to receive result of conversion
  @return nothing
 */

void
RawD301ToCustomTrkPt (Raw_D301_Trk_Point_Type * from, Custom_Trk_Point_Type * to)
{
    LEtoHOST32 ((byte *) & (to->posn.lat), from->lat);	   /* latitude */
    LEtoHOST32 ((byte *) & (to->posn.lon), from->lon);	   /* longitude */
    LEtoHOST32 ((byte *) & (to->time), from->time);		   /* time */
    LEtoHOST32 ((byte *) & (to->alt), from->alt);		   /* altitude */
    to->new_trk = from->new_trk;
}


/**
   Converts a raw D301 track point to GPilotS compact custom internal format

  @param from pointer on raw D301 track point to convert
  @param to pointer on a Compact_Trk_Point_Type struct to receive result of conversion
  @return nothing
 */

void
RawD301ToCompactTrkPt (Raw_D301_Trk_Point_Type * from, Compact_Trk_Point_Type * to)
{
    LEtoHOST32 ((byte *) & (to->posn.lat), from->lat);	   /* latitude */
    LEtoHOST32 ((byte *) & (to->posn.lon), from->lon);	   /* longitude */
    to->new_trk = from->new_trk;
}


/**
   Converts a raw D300 track point to GPilotS standard custom internal format

@param from pointer on raw D300 track point to convert
@param to pointer on a Custom_Trk_Point_Type struct to receive result of conversion
@return nothing
 */

void
RawD300ToCustomTrkPt (Raw_D300_Trk_Point_Type * from, Custom_Trk_Point_Type * to)
{
    LEtoHOST32 ((byte *) & (to->posn.lat), from->lat);	   /* latitude */
    LEtoHOST32 ((byte *) & (to->posn.lon), from->lon);	   /* longitude */
    LEtoHOST32 ((byte *) & (to->time), from->time);		   /* time */
    to->alt = 1.0e25;
    to->new_trk = from->new_trk;
}


/**
  Converts a raw D300 track point to GPilotS compact custom internal format

  @param from pointer on raw D300 track point to convert
  @param to pointer on a Compact_Trk_Point_Type struct to receive result of conversion
  @return nothing
 */

void
RawD300ToCompactTrkPt (Raw_D300_Trk_Point_Type * from, Compact_Trk_Point_Type * to)
{
    LEtoHOST32 ((byte *) & (to->posn.lat), from->lat);	   /* latitude */
    LEtoHOST32 ((byte *) & (to->posn.lon), from->lon);	   /* longitude */
    to->new_trk = from->new_trk;
}


/**
   Converts all known raw track point format to any internal format

   @param  in_format format of raw track point to convert
   @param  out_format format to convert to
   @param  from pointer on raw track point to convert
   @param  to pointer on buffer to receive converted track point
   @return size of converted track point, 0 if error
 */

UInt16
AnyRawToAnyTrkPt (LocalFormatID in_format, LocalFormatID out_format, GarminObject * from,
		  TrkPts * to)
{

    if (out_format == CompactTrkPt)
      {							   /* format compact */
	  switch (in_format)
	    {
	    case G301Type:
		RawD301ToCompactTrkPt (&(from->d301), &(to->compact));
		break;
	    case G300Type:
		RawD300ToCompactTrkPt (&(from->d300), &(to->compact));
		break;
	    default:
		return (0);
	    }
	  return (sizeof (Compact_Trk_Point_Type));
      }
    else
      {							   /* format standard */
	  switch (in_format)
	    {
	    case G301Type:
		RawD301ToCustomTrkPt (&(from->d301), &(to->custom));
		break;
	    case G300Type:
		RawD300ToCustomTrkPt (&(from->d300), &(to->custom));
		break;
	    default:
		return (0);
	    }
	  return (sizeof (Custom_Trk_Point_Type));
      }
    return (0);
}							   /* AnyRawToAnyTrkPt */


/**
   Converts a GPilotS' standard custom format track point to a raw D301 one

  @param from pointer on a Custom_Trk_Point_Type struct to convert
  @param to pointer on raw D301 track point to receive result of conversion
  @return nothing
 */

void
CustomToRawD301TrkPt (Custom_Trk_Point_Type * from, Raw_D301_Trk_Point_Type * to)
{
    MemSet (to, sizeof (Raw_D301_Trk_Point_Type), 0);
    HOSTtoLE32 (to->lat, (byte *) & (from->posn.lat));
    HOSTtoLE32 (to->lon, (byte *) & (from->posn.lon));
    HOSTtoLE32 (to->time, (byte *) & (from->time));
    HOSTtoLE32 (to->alt, (byte *) & (from->alt));
    to->new_trk = from->new_trk;
}


/**
   Converts a GPilotS' compact format track point to a raw D301 one

  @param from pointer on a Compact_Trk_Point_Type struct to convert
  @param to pointer on raw D301 track point to receive result of conversion
  @return nothing
 */

void
CompactToRawD301TrkPt (Compact_Trk_Point_Type * from, Raw_D301_Trk_Point_Type * to)
{
    MemSet (to, sizeof (Raw_D301_Trk_Point_Type), 0);
    HOSTtoLE32 (to->lat, (byte *) & (from->posn.lat));
    HOSTtoLE32 (to->lon, (byte *) & (from->posn.lon));
    to->new_trk = from->new_trk;
}


/**
   Converts a GPilotS'  custom format track point to a raw D300 one

  @param from pointer on a Custom_Trk_Point_Type struct to convert
  @param to pointer on raw D300 track point to receive result of conversion
  @return nothing
*/

void
CustomToRawD300TrkPt (Custom_Trk_Point_Type * from, Raw_D300_Trk_Point_Type * to)
{
    //  MemSet(to, sizeof(Raw_D301_Trk_Point_Type),0);
    HOSTtoLE32 (to->lat, (byte *) & (from->posn.lat));
    HOSTtoLE32 (to->lon, (byte *) & (from->posn.lon));
    HOSTtoLE32 (to->time, (byte *) & (from->time));
    to->new_trk = from->new_trk;
}


/**
   Converts a GPilotS' compact format track point to a raw D300 one

  @param from pointer on a Compact_Trk_Point_Type struct to convert
  @param to pointer on raw D300 track point to receive result of conversion
  @return nothing
*/

void
CompactToRawD300TrkPt (Compact_Trk_Point_Type * from, Raw_D300_Trk_Point_Type * to)
{
    MemSet (to, sizeof (Raw_D301_Trk_Point_Type), 0);
    HOSTtoLE32 (to->lat, (byte *) & (from->posn.lat));
    HOSTtoLE32 (to->lon, (byte *) & (from->posn.lon));
    to->new_trk = from->new_trk;
}


/**
   Converts a GPilotS' custom track header to a D310 one

  @param  trkhdr pointer on an object containing a custom track header
  @param d310 pointer on a raw D310 track header
  @return nothing
*/

void
CustomTrkHdrToD310 (ObjectPtr trkhdr, D310_Trk_Hdr_Type * d310)
{
    d310->dspl = trkhdr->d.CustTrkHdr.dspl;
    d310->color = trkhdr->d.CustTrkHdr.color;
    StrNCopySafe (d310->trk_ident, trkhdr->d.CustTrkHdr.name, sizeof (d310->trk_ident));
}


/**
   Swap of 4 bytes. Used to convert from x86 long and float to 68xxx long and float

   @param pto pointer to destination long or float
   @param pfrom pointer on source long or float
@return
 */


		      // *********  miscellaneous :

#ifdef __m68000


/**
   Converts a  32 bits word from Little Endian to local endian


@param pto pointer on where to put result
@param pfrom pointer on source data
@return nothing
*/

static void
LEtoHOST32 (byte * pto, byte * pfrom)       /* swap of 4 bytes */
{
    pto += 3;
    *(pto--) = *(pfrom++);
    *(pto--) = *(pfrom++);
    *(pto--) = *(pfrom++);
    *(pto--) = *(pfrom++);
}


/**
   Converts a 16 bits word from Little Endian to local endian

   @param pto pointer to destination result
   @param pfrom pointer on source data
@return
 */

static void
LEtoHOST16 (byte * pto, byte * pfrom)       /* swap of 2 bytes */
{
    *(pto + 1) = *(pfrom++);
    *pto = *pfrom;
}

#else  // i386

static void
LEtoHOST32 (byte * pto, byte * pfrom)       /* no-swap of 4 bytes */
{
  UInt32 * fromp= (UInt32 *) pfrom;
  UInt32 * top = (UInt32 *)pto;
  *top=*fromp;
}


/**
   Swap of 2 bytes. Used to convert from x86 int to 68xxx int

   @param pto pointer to destination int
   @param pfrom pointer on source int
@return
 */

static void
LEtoHOST16 (byte * pto, byte * pfrom)       /* no-swap of 2 bytes */
{
    *pto = *(pfrom++);
    *(pto + 1) = *pfrom;
}

#endif

/**
   Converts a D103 symbol type into a Symbol_Type one

   @param  smbl value of symbol type
   @return the corresponding Symbol_Type value
 */

Symbol_Type
smblToSymbolType (byte smbl)
{
    return (smbl2ST[smbl]);
}


/**
   Converts a Symbol_Type into a D103 symbol type.
 A default value is returned if the source value has no corresponding value.

   @param  smbl symbol value to convert
   @return 
 */

byte
SymbolTypeTosmbl (Symbol_Type smbl)
{
    Int16 i;

    for (i = 0; i < 16; i++)
	if (smbl == smbl2ST[i])
	    return (i);
    return (smbl_dot);
}
